/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/**
 *  \file stat.c Functions used for sending/receiving the status of para_server
 *  and para_audiod.
 */


#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "close_on_fork.h"
#include "list.h"
#include "error.h"
#include "string.h"
#include "fd.h"

/** The maximal number of simultaneous connections. */
#define MAX_STAT_CLIENTS 50

/**
 * Describes a status client of para_audiod.
 *
 * There's one such structure per audiod client that sent the 'stat' command.
 *
 * A status client is identified by its file descriptor.  para_audiod
 * keeps a list of connected status clients.
 */
struct stat_client {
	/** The stat client's file descriptor. */
	int fd;
	/** Bitmask of those status items the client is interested in. */
	uint64_t item_mask;
	/** Its entry in the list of stat clients. */
	struct list_head node;
};

static struct list_head client_list;
static int initialized;
static int num_clients;

/** The list of all status items used by para_{server,audiod,gui}. */
const char *status_item_list[] = {STATUS_ITEM_ARRAY};

static void dump_stat_client_list(void)
{
	struct stat_client *sc;

	if (!initialized)
		return;
	list_for_each_entry(sc, &client_list, node)
		PARA_INFO_LOG("stat client on fd %d\n", sc->fd);
}
/**
 * Add a status client to the list.
 *
 * \param fd The file descriptor of the client.
 * \param mask Bitfield of status items for this client.
 *
 * Only those status items having the bit set in \a mask will be
 * sent to the client.
 *
 * \return Positive value on success, or -E_TOO_MANY_CLIENTS if
 * the number of connected clients exceeds #MAX_STAT_CLIENTS.
 */
int stat_client_add(int fd, uint64_t mask)
{
	struct stat_client *new_client;

	if (num_clients >= MAX_STAT_CLIENTS) {
		PARA_ERROR_LOG("maximal number of stat clients (%d) exceeded\n",
			MAX_STAT_CLIENTS);
		return -E_TOO_MANY_CLIENTS;
	}
	if (!initialized) {
		INIT_LIST_HEAD(&client_list);
		initialized = 1;
	}
	PARA_INFO_LOG("adding client on fd %d\n", fd);
	new_client = para_malloc(sizeof(struct stat_client));
	new_client->fd = fd;
	new_client->item_mask = mask;
	para_list_add(&new_client->node, &client_list);
	dump_stat_client_list();
	num_clients++;
	return 1;
}
/**
 * Write a message to all connected status clients.
 *
 * \param msg A \p NULL terminated buffer.
 * \param itemnum The number of the status item of \a msg.
 *
 * On write errors, remove the status client from the client list and close its
 * file descriptor.
 */
void stat_client_write(const char *msg, int itemnum)
{
	struct stat_client *sc, *tmp;
	size_t len = strlen(msg);
	const uint64_t one = 1;

	if (!initialized || !len)
		return;
	list_for_each_entry_safe(sc, tmp, &client_list, node) {
		int fd = sc->fd, ret;

		if (!((one << itemnum) & sc->item_mask))
			continue;
		if (write_ok(fd) > 0) {
			ret = write(fd, msg, len);
	//		PARA_DEBUG_LOG("dumped %s to fd %d, ret = %d\n", msg, fd, ret);
			if (ret == len)
				continue;
		}
		/* write error or fd not ready for writing */
		close(fd);
		num_clients--;
		PARA_INFO_LOG("deleting client on fd %d\n", fd);
		list_del(&sc->node);
		free(sc);
		dump_stat_client_list();
	}
//	if (num_clients)
//		PARA_DEBUG_LOG("%d client(s)\n", num_clients);
}

/**
 * Check if string is a known status item.
 *
 * \param item Buffer containing the text to check.
 *
 * \return If \a item is a valid status item, the number of that status item is
 * returned. Otherwise, this function returns \p -E_UNKNOWN_STAT_ITEM.
 */
int stat_item_valid(const char *item)
{
	int i;
	if (!item || !*item) {
		PARA_ERROR_LOG("%s\n", "no item");
		return -E_UNKNOWN_STAT_ITEM;
	}
	FOR_EACH_STATUS_ITEM(i)
		if (!strcmp(status_item_list[i], item))
			return i;
	PARA_ERROR_LOG("invalid stat item: %s\n", item);
	return -E_UNKNOWN_STAT_ITEM;
}

/**
 * Check if line starts with known status item.
 *
 * \param line Buffer containing the line.
 *
 * \return If the beginning of \a line matches any paraslash status item and is
 * followed by a colon, the number of that status item is returned. Otherwise,
 * this function returns \p -E_UNKNOWN_STAT_ITEM.
 */
int stat_line_valid(const char *line)
{
	int i;
	size_t line_len;

	if (!line || !*line)
		return -E_UNKNOWN_STAT_ITEM;
	line_len = strlen(line);
	FOR_EACH_STATUS_ITEM(i) {
		const char *s = status_item_list[i];
		size_t item_len = strlen(s);

		if (line_len > item_len && line[item_len] == ':' &&
				!strncmp(line, s, item_len))
			return i;
	}
	return -E_UNKNOWN_STAT_ITEM;
}

